﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;
using SimpleJSON;

namespace JustAnotherUser {
    // contains the information that the injection will perform
    public class MorphInjectionLoad : MVRScript {
        private IDictionary<DAZCharacterSelector, List<DAZMorph>> _morphs;

        private JSONStorableBool _emptyStorable,
                                _unlimitedStorable;
        private JSONStorableFloat _durationStorable;

        private JSONStorableStringChooser _morphChooseListIncrement,
                                        _morphChooseListSet;

        private IDictionary<string, JSONStorableFloat> _morphIncrement, // all the morphs that it should modify and its increment
                                                       _morphSet;       // all the morphs that it should modify and its final value

        private static readonly string VERSION = "1.0";

        public bool isEmpty { get { return this._emptyStorable.val; } }

        public float duration { get { return this._durationStorable.val; } }

        public IEnumerable<string> morphNames {
            get {
                HashSet<string> r = new HashSet<string>();
                foreach (List<DAZMorph> e in this._morphs.Values) {
                    foreach (DAZMorph morph in e) r.Add(morph.displayName);
                }
                return r;
            }
        }

        public override void Init() {
            // plugin VaM GUI description
            pluginLabelJSON.val = "MorphInjection [Load] v" + VERSION;

            this._morphs = new Dictionary<DAZCharacterSelector, List<DAZMorph>>();
            this._morphIncrement = new Dictionary<string, JSONStorableFloat>();
            this._morphSet = new Dictionary<string, JSONStorableFloat>();
        }

        // Runs once when plugin loads (after Init)
        protected void Start() {
            // enable Morphs listeners
            SuperController.singleton.onAtomAddedHandlers += (Atom atom) => {
                if (containingAtom.type == "Person" && atom.GetComponentInChildren<DAZCharacterSelector>() != null) {
                    // new person added; resync needed
                    ResyncMorphs();
                }
            };

            SuperController.singleton.onAtomRemovedHandlers += (Atom atom) => {
                if (containingAtom.type == "Person" && atom.GetComponentInChildren<DAZCharacterSelector>() != null) {
                    // new person added; resync needed
                    ResyncMorphs();
                }
            };

            SuperController.singleton.onBeforeSceneSaveHandlers += () => {
                // to save the morphSet (that can be 0; the default value) we need to make sure its value is not the default one
                foreach (JSONStorableFloat e in this._morphSet.Values) {
                    if (e.val == e.defaultVal) e.defaultVal = 0.01f;
                }
            };

            SuperController.singleton.onSceneSavedHandlers += () => {
                // once is saved, we can return to normal
                foreach (JSONStorableFloat e in this._morphSet.Values) e.defaultVal = 0f;
            };

            ResyncMorphs();

            this._unlimitedStorable = new JSONStorableBool("unlimited", false, (bool val) => {
                if (val && this._emptyStorable.val == true) this._emptyStorable.val = false; 
            });
            RegisterBool(this._unlimitedStorable);
            CreateToggle(this._unlimitedStorable); // create it on the right

            this._emptyStorable = new JSONStorableBool("empty", false, (bool val) => {
                // TODO change texture
            });
            RegisterBool(this._emptyStorable);
            CreateToggle(this._emptyStorable);

            // GUI
            this._durationStorable = new JSONStorableFloat("duration", 10.0f, 0.1f, 120.0f);
            RegisterFloat(this._durationStorable);
            CreateSlider(this._durationStorable, true); // create on the right

            this._morphChooseListIncrement = new JSONStorableStringChooser("increment", this.morphNames.ToList(), "", "Add morph to increment", (string morphName) => {
                if (this._morphIncrement.Keys.Contains(morphName)) {
                    SuperController.LogMessage(morphName + " already added");
                    return;
                }

                // TODO remove from the other list
                SuperController.LogMessage("Added " + morphName);
                AddMorph(morphName, true);
            });
            // we don't care about this information
            var linkPopup = CreateFilterablePopup(this._morphChooseListIncrement);
            linkPopup.popupPanelHeight = 600f;

            this._morphChooseListSet = new JSONStorableStringChooser("set", this._morphChooseListIncrement.choices, "", "Add morph to set", (string morphName) => {
                if (this._morphSet.Keys.Contains(morphName)) {
                    SuperController.LogMessage(morphName + " already added");
                    return;
                }

                // TODO remove from the other list
                SuperController.LogMessage("Added " + morphName);
                AddMorph(morphName, false);
            });
            // we don't care about this information
            linkPopup = CreateFilterablePopup(this._morphChooseListSet, true);
            linkPopup.popupPanelHeight = 600f;

            LoadJson(GetPluginJsonFromSave());
        }

        // set as empty, unless it's unlimited
        public void UseLoad() {
            if (!this._unlimitedStorable.val) this._emptyStorable.val = true;
        }

        public List<DAZMorph> getAffectedMorphs(DAZCharacterSelector character) {
            List<DAZMorph> r = new List<DAZMorph>();

            // add increment morphs
            foreach (string morph in this._morphIncrement.Keys) {
                DAZMorph val = FindMorphByName(this._morphs[character], morph);
                if (val != null) r.Add(val);
            }

            // add set morphs
            foreach (string morph in this._morphSet.Keys) {
                DAZMorph val = FindMorphByName(this._morphs[character], morph);
                if (val != null) r.Add(val);
            }

            return r;
        }

        // final morph value after 1 animation ends
        public float getMorphFinalValue(DAZMorph morph) {
            if (this._morphSet.Keys.Contains(morph.displayName)) {
                // set
                return this._morphSet[morph.displayName].val;
            }
            else {
                // increment
                float increment = 0f;
                try {
                    increment = this._morphIncrement[morph.displayName].val;
                } catch (Exception ex) { /* maybe in any list? */ }
                return morph.appliedValue + increment;
            }
        }

        public void ResyncMorphs() {
            this._morphs.Clear();

            foreach (Atom atom in SuperController.singleton.GetAtoms()) {
                DAZCharacterSelector characterSelector;
                if (atom.type != "Person" || (characterSelector = atom.GetComponentInChildren<DAZCharacterSelector>()) == null) continue;

                List<DAZMorph> morphs = new List<DAZMorph>();
                ScanBank(characterSelector.morphBank1, morphs); // @author https://github.com/ProjectCanyon/morph-merger/blob/master/MorphMerger.cs
                this._morphs.Add(characterSelector, morphs);
            }

            if (this._morphChooseListIncrement != null) {
                this._morphChooseListIncrement.choices = this.morphNames.ToList();
                this._morphChooseListSet.choices = this._morphChooseListIncrement.choices;
            }
        }

        private void AddMorph(string morphName, bool increment = true, float value = 0.0f) {
            JSONStorableFloat jsonFloat = new JSONStorableFloat((increment ? "Inc#" : "Set#") + morphName, value, -1.0f, 1.0f);
            RegisterFloat(jsonFloat);
            CreateSlider(jsonFloat, !increment);

            if (increment) this._morphIncrement.Add(morphName, jsonFloat);
            else this._morphSet.Add(morphName, jsonFloat);
        }

        private void LoadJson(JSONNode node) {
            if (node == null) return;

            foreach (string entry in node.AsObject.Keys) {
                switch (entry) {
                    case "duration":
                        this._durationStorable.val = node[entry].AsFloat;
                        break;

                    case "unlimited":
                        this._unlimitedStorable.val = node[entry].AsBool;
                        break;

                    case "empty":
                        this._emptyStorable.val = node[entry].AsBool;
                        break;

                    case "id":
                    case "pluginLabel":
                        break; // ignore

                    default:
                        // tracked morphs
                        try {
                            bool increment = entry.StartsWith("Inc#");
                            if (!increment && !entry.StartsWith("Set#")) break; // I don't know what this property is

                            AddMorph(entry.Remove(0, 4), increment, node[entry].AsFloat);
                        } catch (Exception ex) { }
                        break;
                }
            }
        }

        // @author https://raw.githubusercontent.com/ChrisTopherTa54321/VamScripts/master/FloatMultiParamRandomizer.cs
        public JSONNode GetPluginJsonFromSave() {
            foreach (JSONNode atoms in SuperController.singleton.loadJson["atoms"].AsArray) {
                if (!atoms["id"].Value.Equals(containingAtom.name)) continue;

                foreach (JSONNode storable in atoms["storables"].AsArray) {
                    if (storable["id"].Value == this.storeId) {
                        return storable;
                    }
                }
            }

            return null;
        }

        private void ScanBank(DAZMorphBank bank, List<DAZMorph> morphs) { // TODO only morph (not morph & pose)
            if (bank == null) return;

            foreach (DAZMorph morph in bank.morphs) {
                if (!morph.visible) continue;

                morphs.Add(morph);
            }
        }

        private DAZMorph FindMorphByName(List<DAZMorph> morphs, string name) {
            foreach (DAZMorph morph in morphs) {
                if (!morph.displayName.Equals(name)) continue;

                return morph;
            }

            return null; // not found
        }
    }
}